/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file flexdyn.c
 */

#include <stdlib.h>

#include "kernel.h"
#include "omemcpy.h"
#include "extras.h"

static _kernel_swi_regs regs;

typedef struct flex_block {
  struct flex_block * next;
  char * * back_ptr;
  int size;
  char data[1];
} flex_block;

#define WSE 0x600ec
#define ODA 0x20066
#define OCD 0x2002a

#define init_alloc (0)

static flex_block * first = 0;

static int allocated = 0, memory_free = 0, memory_used = 0;
static flex_block * last = 0;
static int dynamic_area = 0, dynamic_start = 0;

static void flexdyn_close(void) {
  if(dynamic_area) {
    regs.r[0] = 1;
    regs.r[1] = dynamic_area;
    if(_kernel_swi(ODA, & regs, & regs) == 0) {
      dynamic_area = 0;
    }
  }
}

void flexdyn_init(int max, char * name)
{
  if(dynamic_area) return;
  regs.r[0] = 0;
  regs.r[1] = -1;
  regs.r[2] = 0;
  regs.r[3] = -1;
  regs.r[4] = 0x80;
  regs.r[5] = max;
  regs.r[6] = 0;
  regs.r[7] = 0;
  regs.r[8] = (int)name;
  if(_kernel_swi(ODA, & regs, & regs)) {
    werr(TRUE, "Unable to create dynamic area");
  }
  dynamic_area = regs.r[1];
  dynamic_start = regs.r[3];
  memory_free = allocated = 0;
  atexit(flexdyn_close);
}

int flexdyn_max(void)
{
  regs.r[0] = -1;
  regs.r[1] = -1;
  _kernel_swi(WSE, & regs, & regs);
  return regs.r[2] + memory_free - 48;
}

static int allocate_more(int size)
{
  regs.r[0] = -1;
  regs.r[1] = -1;
  _kernel_swi(WSE, & regs, & regs);
  if(regs.r[2] < size + 12)
    return 0;

  regs.r[0] = dynamic_area;
  regs.r[1] = size + 12;
  _kernel_swi(OCD, & regs, & regs);

  regs.r[0] = 2;
  regs.r[1] = dynamic_area;
  _kernel_swi(ODA, & regs, & regs);
  regs.r[0] = regs.r[2];

  if(regs.r[0] < init_alloc + allocated + size + 12)
    {
      regs.r[0] = dynamic_area;
      regs.r[1] = -(size + 12);
      _kernel_swi(OCD, & regs, & regs);
      return 0;
    }
  allocated = regs.r[0] - init_alloc;
  memory_free = allocated - memory_used;
  return 1;
}

int flexdyn_alloc(char * * b, int size)
{
  flex_block * next;

  size = (size + 3) & 0xfffffffc;

  if(memory_free < size + 28)
    if(allocate_more(size - memory_free + 28) == 0)
      return 0;

  if(first == 0)
    first = last = (flex_block *)(dynamic_start);

  memory_free -= size + 12;
  memory_used += size + 12;
  * b = & last -> data[0];
  last -> back_ptr = b;
  last -> size = size;
  next = (flex_block *)((int)last + size + 12);
  last -> next = next;
  last = next;
  last -> next = 0;
  last -> size = 0;
  last -> back_ptr = 0;
  return 1;
}

int flexdyn_allocmax(char * * b)
{
  flex_block * next;
  int size;

  size = flexdyn_max();

  if(memory_free < size + 12)
    if(allocate_more(size - memory_free + 12) == 0) {
      size = flexdyn_max();
      return flexdyn_alloc(b, size);
    }

  if(first == 0)
    first = last = (flex_block *)(init_alloc);

  memory_free -= size + 12;
  memory_used += size + 12;
  * b = & last -> data[0];
  last -> back_ptr = b;
  last -> size = size;
  next = (flex_block *)((int)last + size + 12);
  last -> next = next;
  last = next;
  last -> next = 0;
  last -> size = 0;
  last -> back_ptr = 0;
  return 1;
}

static void reshuffle_used(void)
{
  regs.r[0] = dynamic_area;
  regs.r[1] = memory_used - allocated;
  _kernel_swi(OCD, & regs, & regs);
  regs.r[0] = 2;
  regs.r[1] = dynamic_area;
  _kernel_swi(ODA, & regs, & regs);
  regs.r[0] = regs.r[2];
  allocated = regs.r[0] - init_alloc;
  memory_free = allocated - memory_used;
  if(allocated == 0)
    first = last = 0;
}

static void change_pointers(flex_block * fl, int difference)
{
  while(fl != last) {
    fl -> next = (flex_block *)((int)fl -> next + difference);
    fl = fl -> next;
    if(fl != last)
      * fl -> back_ptr = (char *)((int)(* fl -> back_ptr) + difference);
  }
}

static void change_pointersdown(flex_block * fl, int difference)
{
  while(fl != last) {
    fl -> next = (flex_block *)((int)fl -> next - difference);
    * fl -> back_ptr = (char *)((int)(* fl -> back_ptr) - difference);
    fl = fl -> next;
  }
}

void flexdyn_free(char * * b)
{
  flex_block * thisone = (flex_block *)((int)(* b) - 12);
  int difference;

  if(thisone -> back_ptr != b)
    return; /* Whoops ! */
  *b = 0;
  difference = thisone -> size + 12;
  if(thisone -> next != last) {
    omemcpy(thisone, thisone -> next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, thisone, (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last - difference);
    change_pointersdown(thisone, difference);
  } else {
    last = thisone;
  }
  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  memory_used -= difference;
  memory_free += difference;
  reshuffle_used();
}

int flexdyn_extendby(char * * b, int size)
{
  flex_block * thisone = (flex_block *)((int)(* b) - 12);

  size = (size + 3) & 0xfffffffc;

  if(thisone -> back_ptr != b)
    return 0;
  if(size < 0)
    goto decrease_size;

  if(memory_free < size + 12)
    if(allocate_more(size - memory_free + 12) == 0)
      return 0;
  thisone -> size += size;
  memory_used += size;
  memory_free -= size;
  if(thisone -> next != last) {
    omemcpy((void *)((int)thisone -> next + size), thisone->next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, (void *)((int)thisone -> next + size), (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last + size);
    change_pointers(thisone, size);
  } else {
    thisone -> next = last = (flex_block *)((int)last + size);
  }

  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  return 1;

  decrease_size :
    thisone -> size += size;
  memory_used += size;
  memory_free -= size;
  if(thisone -> next != last) {
    omemcpy((void *)((int)thisone -> next + size), thisone -> next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, (void *)((int)thisone -> next + size), (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last + size);
    change_pointers(thisone, size);
  } else {
    thisone -> next = last = (flex_block *)((int)last + size);
  }

  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  reshuffle_used();

  return 1;
}

int flexdyn_size(char * * a)
{
  flex_block * af = (flex_block *)((int)(* a) - 12);

  if(af -> back_ptr != a)
    return 0;
  return af -> size;
}

int flexdyn_extend(char * * a, int to)
{
  return flexdyn_extendby(a, to - flexdyn_size(a));
}

void flexdyn_swap(char * * a, char * * b)
{
  flex_block * af = (flex_block *)((int)(* a) - 12), * bf = (flex_block *)((int)(* b) - 12);
  char * * c = a;

  if((af -> back_ptr != a) || (bf -> back_ptr != b))
    return;

  af -> back_ptr = b;
  bf -> back_ptr = c;
  *c = &bf -> data[0];
  *b = &af -> data[0];
}

void flexdyn_move(char * * a, char * * b)
{
  flex_block * af = (flex_block *)((int)(* a) - 12);
  if(af -> back_ptr != a) return;
  af -> back_ptr = b;
  *a = 0;
  *b = &af -> data[0];
}
